/* 2D and 3D modified kriging
   developed by dudley.benton@gmail.com
*/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

void Abort(int line,char*format,...)
  {
  va_list arg_marker;
  va_start(arg_marker,format);
  fprintf(stderr,"\nerror %i: ",line);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

int nint(double d)
  {
  if(d>0.)
    return((int)(d+0.5));
  if(d<0.)
    return((int)(d-0.5));
  return(0);
  }

double hypot3(double dX,double dY,double dZ)
  {
  return(sqrt(dX*dX+dY*dY+dZ*dZ));
  }

int dim,Nd,Nx,Ny,Nz;
double*Cd,*Ci,*Ck,Cm,Cx,K,*Xd,*Xi,Xm,Xx,*Yd,*Yi,Ym,Yx,*Zd,*Zi,Zm,Zx;

void ReadData(char*fname)
  {
  char bufr[128];
  int i,j,k,l;
  double C,X,Y,Z;
  FILE*fp;
  fprintf(stderr,"reading data: %s\n",fname);
  if((fp=fopen(fname,"rt"))==NULL)
    Abort(__LINE__,"can't open file");
  l=i=j=0;
  while(fgets(bufr,sizeof(bufr),fp))
    {
    l++;
    k=sscanf(bufr,"%lf%*[, \t]%lf%*[, \t]%lf%*[, \t]%lf",&X,&Y,&Z,&C);
    if(k==3)
      i++;
    else if(k==4)
      j++;
   }
  fprintf(stderr,"%i lines read\n",l);
  if(i>3*j)
    {
    dim=2;
    Nd=i;
    fprintf(stderr,"data is 2D (X,Y,C)\n");
    }
  else if(j>3*i)
    {
    dim=3;
    Nd=j;
    fprintf(stderr,"data is 3D (X,Y,Z,C)\n");
    }
  else
    Abort(__LINE__,"data is not consistently 2D (X,Y,C) or 3D (X,Y,Z,C)");
  fprintf(stderr,"%i points found\n",Nd);
  if(Nd<3*dim)
    Abort(__LINE__,"less than %i points",3*dim);
  if((Cd=calloc(Nd*(dim+2),sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory");
  Ck=Cd+Nd;
  Xd=Ck+Nd;
  Yd=Xd+Nd;
  if(dim==3)
    Zd=Yd+Nd;
  rewind(fp);
  Cm=Xm=Ym=Zm=DBL_MAX;
  Cx=Xx=Yx=Zx=-DBL_MAX;
  Nd=0;
  while(fgets(bufr,sizeof(bufr),fp))
    {
    if(sscanf(bufr,"%lf%*[, \t]%lf%*[, \t]%lf%*[, \t]%lf",&X,&Y,&Z,&C)==dim+1)
      {
      Xd[Nd]=X;
      Yd[Nd]=Y;
      if(X<Xm)
        Xm=X;
      if(X>Xx)
        Xx=X;
      if(Y<Ym)
        Ym=Y;
      if(Y>Yx)
        Yx=Y;
      if(dim<3)
        C=Z;
      else
        {
        if(Z<Zm)
          Zm=Z;
        if(Z>Zx)
          Zx=Z;
        Zd[Nd]=Z;
        }
      Cd[Nd]=C;
      if(C<Cm)
        Cm=C;
      if(C>Cx)
        Cx=C;
      Nd++;
      }
    }
  fclose(fp);
  fprintf(stderr,"%lG<X<%lG\n",Xm,Xx);
  fprintf(stderr,"%lG<Y<%lG\n",Ym,Yx);
  if(dim>2)
    fprintf(stderr,"%lG<Z<%lG\n",Zm,Zx);
  fprintf(stderr,"%lG<C<%lG\n",Cm,Cx);
  }

void Kriging()
  {
  int c,d,w,x,y,z;
  double A,B,*Bd,C,expand=0.05,R;
  if(dim>2)
    K=hypot3(Xx-Xm,Yx-Ym,Zx-Zm)/(2.*M_PI);
  else
    K=hypot(Xx-Xm,Yx-Ym)/(2.*M_PI);
  fprintf(stderr,"characteristic length=%lG\n",K);
  if(Nx==0||Ny==0||Nz==0)
    {
    Nx=max(20,min(80,nint(60*(Xx-Xm)/hypot(Xx-Xm,Yx-Ym))));
    Ny=max(20,min(80,nint(60*(Yx-Ym)/hypot(Xx-Xm,Yx-Ym))));
    if(dim>2)
      Nz=max(20,(Nx+Ny)/2);
    else
      Nz=1;
    }
  if(dim>2)
    fprintf(stderr,"Nx=%i, Ny=%i, Nz=%i\n",Nx,Ny,Nz);
  else
    fprintf(stderr,"Nx=%i, Ny=%i\n",Nx,Ny);
  if((Ci=calloc(Nx*Ny*Nz+Nx+Ny+Nz,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory");
  Xi=Ci+Nx*Ny*Nz;
  Yi=Xi+Nx;
  Zi=Yi+Ny;
  fprintf(stderr,"expanding domain by %lG%%\n",expand*100.);
  A=Xx-Xm;
  B=(Xm+Xx)/2.;
  Xm=B-A*(1.+expand)/2.;
  Xx=B+A*(1.+expand)/2.;
  fprintf(stderr,"%lG<X<%lG\n",Xm,Xx);
  A=Yx-Ym;
  B=(Ym+Yx)/2.;
  Ym=B-A*(1.+expand)/2.;
  Yx=B+A*(1.+expand)/2.;
  fprintf(stderr,"%lG<Y<%lG\n",Ym,Yx);
  if(dim>2)
    {
    A=Zx-Zm;
    B=(Zm+Zx)/2.;
    Zm=B-A*(1.+expand)/2.;
    Zx=B+A*(1.+expand)/2.;
    fprintf(stderr,"%lG<Z<%lG\n",Zm,Zx);
    }
  fprintf(stderr,"begin kriging\n");
  if((Bd=calloc(Nd*Nd,sizeof(double)))==NULL)
    Abort(__LINE__,"can't allocate memory");
  for(w=d=0;d<Nd;d++)
    {
    Ck[d]=Cd[d];
    for(c=0;c<Nd;c++,w++)
      {
      if(c-d)
        {
        if(dim>2)
          R=hypot3(Xd[c]-Xd[d],Yd[c]-Yd[d],Zd[c]-Zd[d]);
        else
          R=hypot(Xd[c]-Xd[d],Yd[c]-Yd[d]);
        Bd[w]=exp(-R*R/K/K);
        }
      else
        Bd[w]=1.;
      }
    }
  for(w=0;w<64;w++)
    {
    for(d=0;d<Nd;d++)
      {
      C=Cd[d];
      for(c=0;c<Nd;c++)
        if(c-d)
          C-=Ck[d]*Bd[Nd*d+c];
      Ck[d]=fmax(Ck[d]/2.,fmin(2.*Cd[d],(w*Ck[d]+C)/(w+1)));
      }
    fprintf(stderr,"\r%.0lf%%",(w+1.)*50./64);
    }
  free(Bd);
  if(dim>2)
    {
    for(w=z=0;z<Nz;z++)
      {
      Zi[z]=Zm+z*(Zx-Zm)/(Nz-1);
      for(y=0;y<Ny;y++)
        {
        Yi[y]=Ym+y*(Yx-Ym)/(Ny-1);
        for(x=0;x<Nx;x++)
          {
          Xi[x]=Xm+x*(Xx-Xm)/(Nx-1);
          for(C=d=0;d<Nd;d++)
            {
            R=hypot3(Xi[x]-Xd[d],Yi[y]-Yd[d],Zi[z]-Zd[d]);
            C+=Ck[d]*exp(-R*R/K/K);
            }
          Ci[w++]=C;
          }
        }
      fprintf(stderr,"\r%i%%",nint(50.+w*50./Nx/Ny/Nz));
      }
    }
  else
    {
    for(w=y=0;y<Ny;y++)
      {
      Yi[y]=Ym+y*(Yx-Ym)/(Ny-1);
      for(x=0;x<Nx;x++)
        {
        Xi[x]=Xm+x*(Xx-Xm)/(Nx-1);
        for(C=d=0;d<Nd;d++)
          {
          R=hypot(Xi[x]-Xd[d],Yi[y]-Yd[d]);
          C+=Ck[d]*exp(-R*R/K/K);
          }
        Ci[w++]=C;
        }
      fprintf(stderr,"\r%i%%",nint(50.+w*50./Nx/Ny));
      }
    }
  fprintf(stderr,"\rdone\n");
  }

void WriteTecplot(char*fname)
  {
  int w,x,y,z;
  FILE*fp;
  fprintf(stderr,"output file: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"# this file can be read by Tecplot\n");
  if(dim>2)
    {
    fprintf(fp,"VARIABLES=\"X\", \"Y\", \"Z\", \"log(C)\"\n");
    fprintf(fp,"ZONE I=%i, J=%i, K=%i\n",Nx,Ny,Nz);
    for(w=z=0;z<Nz;z++)
      for(y=0;y<Ny;y++)
        for(x=0;x<Nx;x++)
          fprintf(fp,"%0.9lG %0.9lG %0.9lG %0.4lG\n",Xi[x],Yi[y],Zi[z],Ci[w++]);
    }
  else
    {
    fprintf(fp,"VARIABLES=\"X\", \"Y\", \"log(C)\"\n");
    fprintf(fp,"ZONE I=%i, J=%i\n",Nx,Ny);
    for(w=y=0;y<Ny;y++)
      for(x=0;x<Nx;x++)
        fprintf(fp,"%0.9lG %0.9lG %0.4lG\n",Xi[x],Yi[y],Ci[w++]);
    }
  fclose(fp);
  }

void WriteTB2(char*fname)
  {
  int x,y,z;
  FILE*fp;
  fprintf(stderr,"output file: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"*this file can be read by TP2\n");
  fprintf(fp,"%i Nx\n",Nx);
  for(x=0;x<Nx;x++)
    fprintf(fp,"%.9lf\n",Xi[x]);
  fprintf(fp,"%i Ny\n",Ny);
  for(y=0;y<Ny;y++)
    fprintf(fp,"%.9lf\n",Yi[y]);
  fprintf(fp,"%i Nx*Ny\n",Nx*Ny);
  for(z=y=0;y<Ny;y++)
    for(x=0;x<Nx;x++)
      fprintf(fp,"%.9lf\n",Ci[z++]);
  fclose(fp);
  }

void WriteTB3(char*fname)
  {
  int w,x,y,z;
  FILE*fp;
  fprintf(stderr,"output file: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"*this file can be read by TP2\n");
  fprintf(fp,"%i Nx\n",Nx);
  for(x=0;x<Nx;x++)
    fprintf(fp,"%.9lf\n",Xi[x]);
  fprintf(fp,"%i Ny\n",Ny);
  for(y=0;y<Ny;y++)
    fprintf(fp,"%.9lf\n",Yi[y]);
  fprintf(fp,"%i Nz\n",Nz);
  for(z=0;z<Nz;z++)
    fprintf(fp,"%.9lf\n",Zi[z]);
  fprintf(fp,"%i Nx*Ny*Nz\n",Nx*Ny*Nz);
  for(w=z=0;z<Nz;z++)
    for(y=0;y<Ny;y++)
      for(x=0;x<Nx;x++)
        fprintf(fp,"%.9lf\n",Ci[w++]);
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  fprintf(stderr,"2D/3D Modified Kriging\n");
  if(argc<2)
    {
    fprintf(stderr,"no data file ... try\nkriging data.xyz 20 20 20\n");
    return(1);
    }
  ReadData(argv[1]);
  if(argc>2&&argc!=2+dim)
    {
    fprintf(stderr,"If you supply Nx,Ny for 2D or Nx,Ny,Nz for 3D\n"
           "you must supply exactly 2 or 3 numbers\n"
           "corresponding to the dimension of the data.\n");
    return(1);
    }
  if(argc>2)
    {
    Nx=atoi(argv[2]);
    Ny=atoi(argv[3]);
    if(dim>2)
      Nz=atoi(argv[4]);
    }
  Kriging();
  WriteTecplot("kriging.dat");
  if(dim>2)
    WriteTB3("kriging.tb3");
  else
    WriteTB2("kriging.tb2");
  return(0);
  }
