/* Gauss-Jordan Elimination (simultaneous linear equations)
   matrix inversion (using Gauss-Jordan elimination)
   Linear Least-Squares with Linear Constraints (using the above)
   by Dudley J. Benton
*/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double powi(double x,int n)
  {
  double p;
  p=1.;
  while(n>0)
    {
    p*=x;
    n--;
    }
  while(n<0)
    {
    p/=x;
    n++;
    }
  return(p);
  }

void MatrixMultiply(double*A,double*B,double*C,int l,int m,int n)
  {
  int i,j,k;
  for(i=0;i<l;i++)
    for(k=0;k<n;k++)
      for(C[n*i+k]=j=0;j<m;j++)
        C[n*i+k]+=A[m*i+j]*B[n*j+k];
  }

void MatrixTranspose(double*A,double*B,int n,int m)
  {
  int i,j;
  for(i=0;i<n;i++)
    for(j=0;j<m;j++)
      B[n*j+i]=A[m*i+j];
  }

int MatrixSolve(double*A,double*B,int n)
  {
  int i,j,k,ip,jp,*pivot;
  double a,p,b;
  if(n<1)
    return(__LINE__);
  if(n==1)
    {
    a=A[0];
    if(fabs(a)<FLT_MIN)
      return(__LINE__);
    B[0]/=a;
    return(0);
    }
  if((pivot=calloc(n,sizeof(int)))==NULL)
    return(__LINE__);
  for(i=0;i<n;i++)
    pivot[i]=i;
  for(k=0;k<n-1;k++)
    {
    ip=k;
    jp=k;
    p=fabs(A[k*(n+1)]);
    for(i=k;i<n;i++)
      {
      for(j=k;j<n;j++)
        {
        a=fabs(A[n*i+j]);
        if(a>p)
          {
          ip=i;
          jp=j;
          p=a;
          }
        }
      }
    if(p<FLT_MIN)
      {
      free(pivot);
      return(__LINE__);
      }
    if(ip!=k)
      {
      b=B[k];
      B[k]=B[ip];
      B[ip]=b;
      for(j=k;j<n;j++)
        {
        a=A[n*k+j];
        A[n*k+j]=A[n*ip+j];
        A[n*ip+j]=a;
        }
      }
    if(jp!=k)
      {
      j=pivot[jp];
      pivot[jp]=pivot[k];
      pivot[k]=j;
      for(i=0;i<n;i++)
        {
        a=A[n*i+k];
        A[n*i+k]=A[n*i+jp];
        A[n*i+jp]=a;
        }
      }
    for(i=k+1;i<n;i++)
      {
      a=A[n*i+k]/A[n*k+k];
      B[i]-=a*B[k];
      for(j=k+1;j<n;j++)
        A[n*i+j]-=a*A[n*k+j];
      }
    }
  a=A[n*n-1];
  p=fabs(a);
  if(p<FLT_MIN)
    {
    free(pivot);
    return(__LINE__);
    }
  B[n-1]/=a;
  for(k=1;k<n;k++)
    {
    i=n-1-k;
    b=0.;
    for(j=i+1;j<n;j++)
      b+=A[n*i+j]*B[j];
    B[i]=(B[i]-b)/A[n*i+i];
    }
  for(i=0;i<n;i++)
    A[i]=B[i];
  for(i=0;i<n;i++)
    B[pivot[i]]=A[i];
  free(pivot);
  return(0);
  }

void TestMatrixSolution(int n)
  {
  int i,j;
  double*A,*B;
  printf("testing %ix%i Gauss-Jordan elimination\n",n,n);
  if((A=calloc(n*n+n,sizeof(double)))==NULL)
    {
    printf("can't allocate memory for arrays\n");
    exit(1);
    }
  B=A+n*n;
  printf("  A      B\n");
  for(i=0;i<n;i++)
    {
    for(j=0;j<n;j++)
      {
      A[n*i+j]=powi(i+1,j);
      printf("%lG ",A[n*i+j]);
      }
    B[i]=exp(i);
    printf("%lG\n",B[i]);
    }
  i=MatrixSolve(A,B,n);
  printf("error code=%i\n",i);
  printf("X=%lG",B[0]);
  for(i=1;i<n;i++)
    printf(",%lG",B[i]);
  printf("\n");
  free(A);
  }

int MatrixInvert(double*A,int n)
  {
  int i,*ipivot,j,*jpivot,k;
  double a,p,r;
  if((ipivot=calloc(2*n,sizeof(int)))==NULL)
    return(__LINE__);
  jpivot=ipivot+n;
  for(k=0;k<n;k++)
    {
    ipivot[k]=k;
    jpivot[k]=k;
    p=fabs(A[n*k+k]);
    for(j=k;j<n;j++)
      {
      for(i=k;i<n;i++)
        {
        a=fabs(A[n*j+i]);
        if(a>p)
          {
          p=a;
          ipivot[k]=i;
          jpivot[k]=j;
          }
        }
      }
    if(p<FLT_MIN)
      {
      free(ipivot);
      return(__LINE__);
      }
    j=ipivot[k];
    if(j!=k)
      {
      for(i=0;i<n;i++)
        {
        a=-A[n*i+k];
        A[n*i+k]=A[n*i+j];
        A[n*i+j]=a;
        }
      }
    i=jpivot[k];
    if(i!=k)
      {
      for(j=0;j<n;j++)
        {
        a=-A[n*k+j];
        A[n*k+j]=A[n*i+j];
        A[n*i+j]=a;
        }
      }
    a=A[n*k+k];
    for(i=0;i<n;i++)
      if(i!=k)
        A[n*k+i]=-A[n*k+i]/a;
    for(i=0;i<n;i++)
      {
      r=A[n*k+i];
      for(j=0;j<n;j++)
        if(i!=k&&j!=k)
          A[n*j+i]+=r*A[n*j+k];
      }
    A[n*k+k]/=a;
    for(j=0;j<n;j++)
      A[n*j+k]/=a;
    }
  for(k=n-2;k>=0;k--)
    {
    i=ipivot[k];
    if(i!=k)
      {
      for(j=0;j<n;j++)
        {
        a=A[n*k+j];
        A[n*k+j]=-A[n*i+j];
        A[n*i+j]=a;
        }
      }
    j=jpivot[k];
    if(j!=k)
      {
      for(i=0;i<n;i++)
        {
        a=A[n*i+k];
        A[n*i+k]=-A[n*i+j];
        A[n*i+j]=a;
        }
      }
    }
  free(ipivot);
  return(0);
  }

void TestMatrixInversion(int n)
  {
  int i,j;
  double*A,*B,*C;
  printf("testing inversion on matrix of order %i\n",n);
  if((A=calloc(n*n*3,sizeof(double)))==NULL)
    {
    printf("can't allocate memory for arrays\n");
    exit(1);
    }
  B=A+n*n;
  C=B+n*n;
  for(i=0;i<n;i++)
    for(j=0;j<n;j++)
      A[n*i+j]=powi(i+1,j);
  memcpy(B,A,n*n*sizeof(double));
  if(MatrixInvert(B,n))
    {
    printf("matrix is singular\n");
    exit(1);
    }
  MatrixMultiply(A,B,C,n,n,n);
  printf("A=\n");
  for(i=0;i<n;i++)
    {
    for(j=0;j<n;j++)
      printf(" %7.4lf",A[n*i+j]);
    printf("\n");
    }
  printf("B=\n");
  for(i=0;i<n;i++)
    {
    for(j=0;j<n;j++)
      printf(" %7.4lf",B[n*i+j]);
    printf("\n");
    }
  printf("C=\n");
  for(i=0;i<n;i++)
    {
    for(j=0;j<n;j++)
      printf(" %7.4lf",C[n*i+j]);
    printf("\n");
    }
  free(A);
  }

int LinearEquationsWithLinearConstraints(double*A,double*B,int m,double*C,double*D,int n,double*X)
  {
  int i,j,l;
  double*AC,*DL;
  l=m+n;
  if((AC=calloc(l*l+l,sizeof(double)))==NULL)
    {
    printf("can't allocate memory for arrays\n");
    exit(1);
    }
  DL=AC+l*l;
  for(i=0;i<m;i++)
    {
    for(j=0;j<m;j++)
      AC[l*i+j]=A[m*i+j];
    DL[i]=B[i];
    }
  for(i=0;i<n;i++)
    {
    for(j=0;j<m;j++)
      {
      AC[l*(m+i)+j  ]=
      AC[l*  j  +m+i]=C[m*i+j];
      }
    DL[m+i]=D[i];
    }
  i=MatrixSolve(AC,DL,l);
  if(i)
    {
    free(AC);
    return(i);
    }
  for(i=0;i<m;i++)
    X[i]=DL[i];
  free(AC);
  return(0);
  }

void TestLinearEquationWithLinearConstraints(int nr,int nt,int nc)
  {
  int c,r,t;
  double*A,*At,*AtA,*AtB,*B,*C,*D,*X,x,y;
  printf("testing linear equation with linear constraints\n");
  if((A=calloc(nt*(nc+2*nr+2+nt)+nc+nr,sizeof(double)))==NULL)
    {
    printf("can't allocate memory for arrays\n");
    exit(1);
    }
  At=A+nr*nt;
  AtA=At+nt*nr;
  AtB=AtA+nt*nt;
  B=AtB+nt;
  C=B+nr;
  D=C+nc*nt;
  X=D+nc;
  for(c=r=0;r<nr;r++)
    {
    x=r;
    y=exp(x);
    for(t=0;t<nt;t++)
      A[nt*r+t]=powi(x,t);
    B[r]=y;
    if(r%(nr/nc)==0&&c<nc)
      {
      for(t=0;t<nt;t++)
        C[nt*c+t]=A[nt*r+t];
      D[c]=B[r];
      c++;
      }
    }
  MatrixTranspose(A,At,nr,nt);
  MatrixMultiply(At,A,AtA,nt,nr,nt);
  MatrixMultiply(At,B,AtB,nt,nr,1);
  r=LinearEquationsWithLinearConstraints(AtA,AtB,nt,C,D,nc,X);
  printf("error code=%i\n",r);
  printf("X=%lG",X[0]);
  for(t=1;t<nt;t++)
    printf(",%lG",X[t]);
  printf("\nx  y  fit\n");
  for(r=0;r<nr;r++)
    {
    x=r;
    y=X[nt-1];
    for(t=nt-2;t>=0;t--)
      y=y*x+X[t];
    printf("%lG %lG %lG\n",x,exp(x),y);
    }
  free(A);
  }

int main(int argc,char**argv,char**envp)
  {
  TestMatrixSolution(3);
  TestMatrixInversion(3);
  TestLinearEquationWithLinearConstraints(7,5,2);
  return(0);
  }
